from ailment.expression import BinaryOp, Const

from .base import PeepholeOptimizationExprBase


class ASubADiv(PeepholeOptimizationExprBase):
    __slots__ = ()

    NAME = "a - a / N => a * (N - 1) / N"
    expr_classes = (BinaryOp, )

    def optimize(self, expr: BinaryOp):

        if expr.op == "Sub" and len(expr.operands) == 2 \
                and isinstance(expr.operands[1], BinaryOp) and expr.operands[1].op == "Div" \
                and isinstance(expr.operands[1].operands[1], Const):
            a = expr.operands[0]
            if expr.operands[1].operands[0].likes(a):
                N = expr.operands[1].operands[1].value
                mul = BinaryOp(expr.idx, "Mul",
                               [a,
                                Const(None, None, N - 1, expr.bits)
                                ],
                               False,
                               **expr.tags)
                div = BinaryOp(expr.operands[1].idx, "Div",
                               [mul,
                                Const(None, None, N, expr.bits, **expr.operands[1].tags)
                               ],
                               False,
                               **expr.operands[1].tags)
                return div

        return None



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("msports.dll")
prototypes = \
    {
        # 
        'ComDBOpen': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["PHComDB"]),
        # 
        'ComDBClose': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["HComDB"]),
        # 
        'ComDBGetCurrentPortUsage': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["HComDB", "Buffer", "BufferSize", "ReportType", "MaxPortsReported"]),
        # 
        'ComDBClaimNextFreePort': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["HComDB", "ComNumber"]),
        # 
        'ComDBClaimPort': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["HComDB", "ComNumber", "ForceClaim", "Forced"]),
        # 
        'ComDBReleasePort': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["HComDB", "ComNumber"]),
        # 
        'ComDBResizeDatabase': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["HComDB", "NewSize"]),
    }

lib.set_prototypes(prototypes)



import math

from ailment.expression import Convert, BinaryOp, Const

from .base import PeepholeOptimizationExprBase


class ConvConstMullAShift(PeepholeOptimizationExprBase):
    __slots__ = ()

    NAME = "Conv(64->32, (N * a) >> 32) => a / N1"
    expr_classes = (Convert, )

    def optimize(self, expr: Convert):

        if expr.from_bits == 64 and expr.to_bits == 32 \
                and isinstance(expr.operand, BinaryOp) and expr.operand.op == "Shr" \
                and isinstance(expr.operand.operands[1], Const) \
                and expr.operand.operands[1].value == 32:

            inner = expr.operand.operands[0]
            if isinstance(inner, BinaryOp) and inner.op == "Mull" and isinstance(inner.operands[0], Const):
                bits = 32
                C = inner.operands[0].value
                X = inner.operands[1]
                V = bits
                ndigits = 5 if V == 32 else 6
                divisor = self._check_divisor(pow(2, V), C, ndigits)
                if divisor is not None:
                    new_const = Const(None, None, divisor, V)
                    new_expr = BinaryOp(inner.idx, 'Div', [X, new_const], inner.signed, **inner.tags)
                    return new_expr

    @staticmethod
    def _check_divisor(a, b, ndigits=6):
        divisor_1 = 1 + (a//b)
        divisor_2 = int(round(a/float(b), ndigits))
        return divisor_1 if divisor_1 == divisor_2 else None



from ailment.expression import BinaryOp, Const

from .base import PeepholeOptimizationExprBase


class ASubADivConstMulConst(PeepholeOptimizationExprBase):
    __slots__ = ()

    NAME = "a - (a / N) * N => a % N"
    expr_classes = (BinaryOp, )

    def optimize(self, expr: BinaryOp):

        if expr.op == "Sub" and len(expr.operands) == 2 \
                and isinstance(expr.operands[1], BinaryOp) and expr.operands[1].op == "Mul" \
                and isinstance(expr.operands[1].operands[1], Const):
            a0 = expr.operands[0]
            op1 = expr.operands[1]
            mul_const = expr.operands[1].operands[1]
            if isinstance(op1.operands[0], BinaryOp) and op1.operands[0].op == "Div" and \
                    isinstance(op1.operands[0].operands[1], Const):
                a1 = op1.operands[0].operands[0]
                div_const = op1.operands[0].operands[1]

                if a0.likes(a1) and mul_const.value == div_const.value:
                    mod = BinaryOp(expr.idx, "DivMod",
                                   [a0,
                                    div_const
                                   ],
                                   False,
                                   **expr.tags)
                    return mod

        return None
